#
# Copyright (c) 2015, Lucas Rentschler (lrentschler@ufm.edu)
#                     Theodore L. Turocy (T.Turocy@uea.ac.uk)
#
# A utility for computing equilibria in two-bidder all-pay auctions
# with interdependent valuations
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
"""Files for Example 1 in paper:
Common-value auction with conditionally independent signals.
"""

import argparse

import sympy

import auction
import enumerator

class TwoSignalCVAuction(auction.ConditionalSignalsAuction):
    """Common value auction, where two signals are possible for
    each realized common value.  If value v_k is realized, then
    the possible signals are s_{k-1} or s_k.  

    The optional parameter 'p' controls the probability that the
    "higher" signal is realized.  The case in the paper looks only
    at the p=1/2 case.
    """
    def __init__(self, values, p=sympy.Rational(1,2)):
        self.values = [ 0 ] + values
        self.p = p
        auction.ConditionalSignalsAuction.__init__(self)
                
    def f(self, m):
        """Probability that values[m] is the value which is realized.
        """
        if m > 0:
            return sympy.Rational(1, len(self.values) - 1)
        else:
            return 0

    def g(self, k, m):
        "Probability of signal s_k given value v_m."
        if m == k:
            return self.p
        elif m == k+1:
            return 1 - self.p
        else:
            return 0

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Solve Example 1 (common values with 2 signals possible per value)')
    parser.add_argument('K', type=int,
                        help='number of values (uniformly spaced on [0, 1])')

    args = parser.parse_args()
    values = [ sympy.Rational(i, args.K)
               for i in xrange(1, args.K+1) ]
    game = TwoSignalCVAuction(values)
    enumerator.enumerate_solutions(game,
                                   enumerator.PathEnumerator(game),
                                   enumerator.EquilibriumList())
