#
# Copyright (c) 2015, Lucas Rentschler (lrentschler@ufm.edu)
#                     Theodore L. Turocy (T.Turocy@uea.ac.uk)
#
# A utility for computing equilibria in two-bidder all-pay auctions
# with interdependent valuations
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
"""Solve Example 3a.
"""

import argparse
import sympy

import auction
import enumerator

class RightWrongAuction(auction.ConditionalSignalsAuction):
    """Auction where signal is 'correct' with probability 'pC', and
    other signals occur with probability 'pW'.
    """
    def __init__(self, values, pC):
        self.values = values
        self.pC = pC
        self.pW = sympy.Rational(1-pC, len(self.values)-1)
        auction.ConditionalSignalsAuction.__init__(self)

    def f(self, m):
        "Probability of value v_m being realized."
        return sympy.Rational(1, len(self.values))

    def g(self, k, m):
        "Probability of signal s_k given value v_m."
        if k == m:  return self.pC
        return self.pW

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Solve game in Example 3a.')
    parser.add_argument('K', type=int,
                        help='number of values (assumes uniform grid)')
    parser.add_argument('pC', type=sympy.Rational,
                        help='probabilty of correct signal')
    
    args = parser.parse_args()
    values = [ sympy.Rational(i, args.K)
               for i in xrange(1, args.K+1) ]
    game = RightWrongAuction(values, args.pC)
    enumerator.enumerate_solutions(game,
                                   enumerator.PathEnumerator(game),
                                   enumerator.EquilibriumList())
